#include "Solver.h"
#include "FormulaTranslator.hh"
#include <set>
#include <ostream>
#include <ppl.hh>

namespace PPL = Parma_Polyhedra_Library;
using PPL::operator+;
using PPL::operator-;
using PPL::operator*;
using PPL::operator<=;
using PPL::operator>=;
using PPL::IO_Operators::operator<<;

#define GENERATE_TRIVIAL_THEORY

Lit FormulaTranslator::getPropositionalVariable(const std::string& var) const
{
  std::map<std::string, Var>::const_iterator it =
    propositionalVariableMap.find(var);
  if (it == propositionalVariableMap.end()) {
    std::cerr << "prop variable <" << var << "> not found" << std::endl;
    std::terminate();
  }

  return Lit(it->second);
}

unsigned FormulaTranslator::getNumericVariable(const std::string& var) const
{
  std::map<std::string, unsigned>::const_iterator it =
    numericVariableMap.find(var);
  if (it == numericVariableMap.end()) {
    std::cerr << "num variable <" << var << "> not found" << std::endl;
    std::terminate();
  }
  unsigned v = it->second;
  assert(v < solver->readLinearDimension());
  return v;
}

void FormulaTranslator::initNumericVariableMap(const Formula& f) {
  std::set<std::string> s;
  f.collectNumericVariables(s);
  for(std::set<std::string>::const_iterator i=s.begin();
      i != s.end();
      ++i) {
    numericVariableMap[*i] = currentNumericVariableIndex++;
    numericVariableRevMap.push_back(*i);
  }
}

void FormulaTranslator::initPropositionalVariableMap(const Formula& f) {
  std::set<std::string> s;
  f.collectPropositionalVariables(s);
  for(std::set<std::string>::const_iterator i=s.begin();
      i != s.end();
      ++i) {
    propositionalVariableMap[*i] = solver->newVar();
  }
}

FormulaTranslator::FormulaTranslator(const Formula& f) :
  currentNumericVariableIndex(0) {
  initNumericVariableMap(f);
  solver = new LinearSMT(currentNumericVariableIndex);
  initPropositionalVariableMap(f);
  goalLiteral = f.compile(*this);
  assert(formulaBindings.begin() == formulaBindings.end());
  assert(exprBindings.begin() == exprBindings.end());
}

FormulaTranslator::~FormulaTranslator() {
  delete solver;
}

void FormulaTranslator::printModel(std::ostream& o) const {
  o << "PROPOSITIONAL VARIABLES" << std::endl;
  for(std::map<std::string, Var>::const_iterator i=propositionalVariableMap.begin(),
	end=propositionalVariableMap.end();
      i != end;
      ++i) {
    o << i->first << " " << solver->model[i->second] << std::endl;
  }
  o << "LINEAR CONSTRAINTS" << std::endl;
  std::set <PPL::Linear_Expression, LinearSMT::Linear_Expression_compare> positiveConstraints;

#ifdef OLD_CONSTRAINT_PRINT
  for(LinearSMT::constraintMap_const_iterator i=solver->constraintMap_begin(), z=solver->constraintMap_end();
      i!=z; ++i) {
    switch(solver->model[i->second].toInt()) {
    case 1:
      positiveConstraints.insert(i->first);
      break;
    case -1:
      o << "(< " << *decompile(i->first) << " 0)" << std::endl;
      break;
    }
  }
  for(std::set <PPL::Linear_Expression, LinearSMT::Linear_Expression_compare>::const_iterator i=positiveConstraints.begin(), z=positiveConstraints.end();
      i != z; ++i) {
    std::set <PPL::Linear_Expression, LinearSMT::Linear_Expression_compare>::iterator negated = positiveConstraints.find(-(*i));
    if (negated != z) {
      positiveConstraints.erase(negated);
      o << "(= " << *decompile(*i) << " 0)" << std::endl;
    } else {
      o << "(>= " << *decompile(*i) << " 0)" << std::endl;
    }
  }
#else
  PPL::Constraint_System cs;
  for(LinearSMT::constraintMap_const_iterator i=solver->constraintMap_begin(), z=solver->constraintMap_end();
      i!=z; ++i) {
    switch(solver->model[i->second].toInt()) {
    case 1:
      cs.insert(i->first >= 0);
      break;
    case -1:
      cs.insert(i->first < 0);
      break;
    }
  }
  PPL::NNC_Polyhedron polyhedron(cs);
  assert(!polyhedron.is_empty());
  const PPL::Constraint_System& csMin = polyhedron.minimized_constraints ();
  for(PPL::Constraint_System::const_iterator i=csMin.begin(), z=csMin.end();
      i != z; ++i) {
    Formula *e = decompile(*i);
    o << *e << std::endl;
    delete e;
  }
#endif
}

Lit PropositionalVariable::compile(FormulaCompiler& compiler) const {
  if (compiler.isFormulaVariableBound(name)) {
    return compiler.readFormulaBinding(name);
  } else {
    return compiler.getPropositionalVariable(name);
  }
}

Lit Not::compile(FormulaCompiler& compiler) const {
  return compiler.mknot(f->compile(compiler));
}

Lit And::compile(FormulaCompiler& compiler) const {
  const_iterator i=begin(), z=end();
  if (i==z) {
    return compiler.mktrue();
  } else {
    Lit last = (*i)->compile(compiler);
    ++i;
    for(; i!=z; ++i) {
      last = compiler.mkand(last, (*i)->compile(compiler));
    }
    return last;
  }
}

Lit Or::compile(FormulaCompiler& compiler) const {
  const_iterator i=begin(), z=end();
  if (i==z) {
    return compiler.mkfalse();
  } else {
    Lit last = (*i)->compile(compiler);
    ++i;
    for(; i!=z; ++i) {
      last = compiler.mkor(last, (*i)->compile(compiler));
    }
    return last;
  }
}

Lit Comparison::compile(FormulaCompiler& compiler) const {
  QLinear_Expression f1 = e1->compile(compiler),
    f2 = e2->compile(compiler);
#if 0
  std::cout << "comparison between " << f1 << " and " << f2 << std::endl;
#endif
  switch (comparisonKind) {
  case LESS:
    return compiler.getLinearConstraintVariable(f1 < f2);
  case LESS_EQ:
    return compiler.getLinearConstraintVariable(f1 <= f2);
  case GREATER:
    return compiler.getLinearConstraintVariable(f1 > f2);
  case GREATER_EQ:
    return compiler.getLinearConstraintVariable(f1 >= f2);
  }
  exit(1);
}

Lit FLet::compile(FormulaCompiler& compiler) const {
  if (compiler.isFormulaVariableBound(boundVariable)) {
    Lit oldBinding = compiler.readFormulaBinding(boundVariable);
    compiler.assignFormulaBinding(boundVariable, binding->compile(compiler));
    Lit x = subFormula->compile(compiler);
    compiler.assignFormulaBinding(boundVariable, oldBinding);
    return x;
  } else {
    compiler.assignFormulaBinding(boundVariable, binding->compile(compiler));
    Lit x = subFormula->compile(compiler);
    compiler.removeFormulaBinding(boundVariable);
    return x;
  }
}

Lit Equality::compile(FormulaCompiler& compiler) const {
  const_iterator prev = begin();
  if (prev==end()) return compiler.mktrue();

  QLinear_Expression prevExpr = (*prev)->compile(compiler);

  const_iterator next = prev; ++next;
  bool first = true;
  Lit output;

  while (next != end()) {
    QLinear_Expression nextExpr = (*next)->compile(compiler);
    Lit eqLit = compiler.getLinearConstraintVariable(prevExpr == nextExpr);

    if (first) output = eqLit;
    else output = compiler.mkand(output, eqLit);

    first = false;
    prev = next;
    prevExpr = nextExpr;
    ++next;
  }
  return output;
}

Lit Iff::compile(FormulaCompiler& compiler) const {
  const_iterator prev = begin();
  if (prev==end()) return compiler.mktrue();

  Lit prevExpr = (*prev)->compile(compiler);

  const_iterator next = prev; ++next;
  bool first = true;
  Lit output;

  while (next != end()) {
    Lit nextExpr = (*next)->compile(compiler);
    Lit eqLit = compiler.mkiff(prevExpr, nextExpr);

    if (first) output = eqLit;
    else output = compiler.mkand(output, eqLit);

    first = false;
    prev = next;
    prevExpr = nextExpr;
    ++next;
  }
  return output;
}

Lit PairwiseDistinct::compile(FormulaCompiler& compiler) const {
  bool first = true;
  Lit output;
  for(const_iterator i=begin(); i!=end(); ++i) {
    QLinear_Expression iExpr = (*i)->compile(compiler);
    
    const_iterator j=i; ++j;
    for(; j!=end(); ++j) {
      QLinear_Expression jExpr = (*j)->compile(compiler);

      Lit eqLit = ~ compiler.getLinearConstraintVariable(iExpr == jExpr);
      
      if (first) output = eqLit;
      else output = compiler.mkand(output, eqLit);

      first = false;
    }
  }
  if (first) {
    return compiler.mktrue();
  } else {
    return output;
  }
}

Formula* FormulaTranslator::decompile(const PPL::Constraint& constraint) const {
  Expression *e = decompile(PPL::Linear_Expression(constraint));
  if (constraint.is_equality()) {
    return new Equality(e, new IntegerConstant(0));
  } else if (constraint.is_strict_inequality()) {
    return new Comparison(e, Comparison::GREATER, new IntegerConstant(0));
  } else {
    return new Comparison(e, Comparison::GREATER_EQ, new IntegerConstant(0));
  }
}

Expression* FormulaTranslator::decompile(const PPL::Linear_Expression& expr) const {
  Sum *o = new Sum();
  o->addExpression(new IntegerConstant(expr.inhomogeneous_term()));
  assert(expr.space_dimension() <= numericVariableRevMap.size());
  for(unsigned i=0; i<expr.space_dimension(); i++) {
    mpz_class coeff = expr.coefficient(PPL::Variable(i));
    if (coeff != 0) {
      if (coeff == 1) {
	o->addExpression(new NumericVariable(numericVariableRevMap[i]));
      } else {
	o->addExpression(new Product(new IntegerConstant(coeff), new NumericVariable(numericVariableRevMap[i])));
      }
    }
  }
  return o;
}

bool FormulaTranslator::solve() {
#ifdef GENERATE_TRIVIAL_THEORY
  solver->generateTrivialTheoryConsequences();
#endif

  vec<Lit> v;
  v.push(goalLiteral);
  return solver->solve(v);
}

void FormulaTranslator::assignFormulaBinding(const std::string& var, Lit x) {
  formulaBindings.insert(std::pair<std::string, Lit>(var, x));
}

void FormulaTranslator::removeFormulaBinding(const std::string& var) {
  formulaBindings.erase(var);
}

bool FormulaTranslator::isFormulaVariableBound(const std::string& var) const {
  return formulaBindings.count(var) > 0;
}

Lit FormulaTranslator::readFormulaBinding(const std::string& var) const {
  return formulaBindings.find(var)->second;
}

void FormulaTranslator::assignExprBinding(const std::string& var, QLinear_Expression e) {
  exprBindings.insert(std::pair<std::string, QLinear_Expression>(var, e));
}

void FormulaTranslator::removeExprBinding(const std::string& var) {
  exprBindings.erase(var);
}

bool FormulaTranslator::isExprVariableBound(const std::string& var) const {
  return exprBindings.count(var) > 0;
}

QLinear_Expression FormulaTranslator::readExprBinding(const std::string& var) const {
  return exprBindings.find(var)->second;
}

Lit ExprLet::compile(FormulaCompiler& compiler) const {
  if (compiler.isExprVariableBound(boundVariable)) {
    QLinear_Expression oldBinding = compiler.readExprBinding(boundVariable);
    compiler.assignExprBinding(boundVariable, binding->compile(compiler));
    Lit x = subFormula->compile(compiler);
    compiler.assignExprBinding(boundVariable, oldBinding);
    return x;
  } else {
    compiler.assignExprBinding(boundVariable, binding->compile(compiler));
    Lit x = subFormula->compile(compiler);
    compiler.removeExprBinding(boundVariable);
    return x;
  }
}

Lit FormulaIfThenElse::compile(FormulaCompiler& compiler) const {
  Lit cv = condition->compile(compiler);
  return compiler.mkor(compiler.mkand(cv, e1->compile(compiler)),
		       compiler.mkand(~cv, e2->compile(compiler)));
}

QLinear_Expression UnaryNeg::compile(FormulaCompiler& compiler) const {
  return - f->compile(compiler);
}

QLinear_Expression NumericVariable::compile(FormulaCompiler& compiler) const {
  if (compiler.isExprVariableBound(name)) {
    return compiler.readExprBinding(name);
  } else {
    return QLinear_Expression(PPL::Variable(compiler.getNumericVariable(name)));
  }
}

QLinear_Expression IntegerConstant::compile(FormulaCompiler& compiler) const {
  return QLinear_Expression(value);
}

QLinear_Expression Sum::compile(FormulaCompiler& compiler) const {
  QLinear_Expression e;
  for(const_iterator i=begin(); i!=end(); ++i) {
    e += (*i)->compile(compiler);
  }
  return e;
}

QLinear_Expression Product::compile(FormulaCompiler& compiler) const {
  QLinear_Expression e(mpz_class(1));
  for(const_iterator i=begin(); i!=end(); ++i) {
    e *= (*i)->compile(compiler);
  }
  return e;
}

QLinear_Expression Quotient::compile(FormulaCompiler& compiler) const {
  return e1->compile(compiler) / e2->compile(compiler);
}

QLinear_Expression ExpressionIfThenElse::compile(FormulaCompiler& compiler) const {
  Lit cLit = condition->compile(compiler);
  PPL::Variable outVar(compiler.getNumericVariable(variableName()));
  QLinear_Expression e1Q = e1->compile(compiler), e2Q = e2->compile(compiler);
  compiler.assertAuxiliary(compiler.mkor(
    compiler.mkand(cLit, compiler.getLinearConstraintVariable(e1Q == QLinear_Expression(outVar))),
    compiler.mkand(compiler.mknot(cLit),
		   compiler.getLinearConstraintVariable(e2Q == QLinear_Expression(outVar)))));
  return QLinear_Expression(outVar);
}
