#include <iostream>
#include <exception>
#include <map>
#include <ppl.hh>
#include <stdlib.h>
#include "LinearSMT.hh"

using namespace PPL::IO_Operators;
using namespace PPL;

static inline unsigned max(unsigned a, unsigned b) {
  return a>=b ? a : b;
}

Variables variables;

bool linear_expression_less (const PPL::Linear_Expression& a, const PPL::Linear_Expression& b) {
  unsigned m = max(a.space_dimension(), b.space_dimension());
  for(unsigned i=0; i<m; i++) {
    Variable& v = variables.get(i);
    if (a.coefficient(v) < b.coefficient(v)) return true;
    if (a.coefficient(v) > b.coefficient(v)) return false;
    }
  if (a.inhomogeneous_term() < b.inhomogeneous_term()) return true;
  return false;
}

bool detect_contradiction(const std::vector<PPL::Constraint>& constraints, std::vector<unsigned> *contradiction);

void LinearSMT::extractContradiction() {
  theoryContradiction.clear();
  if (minimizeContradiction) {
    std::vector <PPL::Constraint> constraints;
    std::vector <Lit> constraintVariables;
    std::vector<unsigned> contradiction;

    for(int i=0; i<trail.size(); i++) {
      if (hasConstraintValue(trail[i])) {
	constraints.push_back(constraintValue(trail[i]));
	constraintVariables.push_back(trail[i]);
      }
    }

    if (! detect_contradiction(constraints, &contradiction)) {
      std::cerr << "contradiction detected but no contradiction in dual" << std::endl;
      std::terminate();
    }

    for(unsigned i=0; i<contradiction.size(); i++) {
      theoryContradiction.push(~ constraintVariables[contradiction[i]]);
    }

#ifdef SMT_DEBUG
    std::cout << "contradiction of " << contradiction.size() <<
      " constraints among " << constraints.size() << std::endl;
#endif
  } else {
    for(int i=0; i<trail.size(); i++) {
      if (hasConstraintValue(trail[i])) {
	theoryContradiction.push(~ trail[i]);
      }
    }
  }
}

void LinearSMT::checkForFinalContradiction() {
#ifdef SMT_DEBUG
      std::cout << "checking final contradiction" << std::endl;
#endif
  if (currentPolyhedron.is_empty()) {
    extractContradiction();
#ifdef SMT_DEBUG
    std::cout << "contradiction reached" << std::endl;
#endif
    throw TheoryContradiction();
  }
}

void LinearSMT::assertLiteral(Lit p) {
#ifdef SMT_DEBUG


    std::cout << "asserting " << var(p) << " " << sign(p) << std::endl;
#endif
    if (hasConstraintValue(p)) {
      Constraint p2 = constraintValue(p);
#ifdef SMT_DEBUG
      std::cout << "begin asserting " << p2 << std::endl;
#endif
      currentPolyhedron.add_constraint(p2);
      if (random() < checkingRandomThreshold) {
        bool empty = currentPolyhedron.is_empty();
#ifdef SMT_DEBUG
        std::cout << "finished asserting" << std::endl;
#endif
        if (empty) {
#ifdef SMT_DEBUG
	  std::cout << "contradiction" << std::endl;
#endif
	  extractContradiction();
	  throw TheoryContradiction();
        }
      }
    }
  }

void LinearSMT::newDecisionLevel () {
#ifdef SMT_DEBUG
  std::cout << "new decision level at " << decisionLevel() << std::endl;
#endif
  polyhedronStack = new Constraint_Polyhedron_stack_element(currentPolyhedron, polyhedronStack);
  SOLVER_SUPERCLASS::newDecisionLevel();
}

void LinearSMT::cancelUntil(int level) {
#ifdef SMT_DEBUG
  std::cout << "cancel until " << level << std::endl;
#endif
  if (decisionLevel() > level) {
    for(int i=0; i<decisionLevel()-(level+1); i++) {
      Constraint_Polyhedron_stack_element *next = polyhedronStack->next;
      delete polyhedronStack;
      polyhedronStack = next;
    }
    currentPolyhedron = polyhedronStack->p;
    Constraint_Polyhedron_stack_element *next = polyhedronStack->next;
    delete polyhedronStack;
    polyhedronStack = next;
  }
  if (level == 0) assert(polyhedronStack == 0);
  SOLVER_SUPERCLASS::cancelUntil(level);
}

LinearSMT::LinearSMT(int dimension) :
  SOLVER_SUPERCLASS(),
  minimizeContradiction(true),
  checkingRandomThreshold((int) (0 * RAND_MAX)),
  linearDimension(dimension),
  universePolyhedron(dimension),
  currentPolyhedron(universePolyhedron),
  polyhedronStack(0) {
  trueVar = newVar();
  vec<Lit> clause;
  clause.push(Lit(trueVar));
  addClause(clause);
}

bool is_zero_form(const Linear_Expression& expr) {
  for(unsigned i=0; i<expr.space_dimension(); ++i) {
    if (expr.coefficient(PPL::Variable(i)) != 0) return false;
  }
  return true;
}

Lit LinearSMT::exprAtom(const Linear_Expression& c) {
  if (is_zero_form(c)) {
    if (c.inhomogeneous_term() >= 0) return mktrue(); else return mkfalse();
  } else {
    std::map <Linear_Expression, Var>::iterator pos = constraintMap.find(c);
    if (pos == constraintMap.end()) {
      Var var=newVar();
#ifdef SMT_DEBUG
      std::cout << "creating new var " << var << " " << linearSemantics.size() << std::endl;
#endif
      constraintMap.insert(pos, std::pair <Linear_Expression, Var>(c, var));
      if (var >= linearSemantics.size()) {
	linearSemantics.growTo(var+1);
      }
      linearSemantics[var] = new Constraint(c >= 0);
      return Lit(var);
    } else {
      return Lit(pos->second);
    }
  }
}

Lit LinearSMT::mkand(Lit a, Lit b) {
#ifndef DO_NOT_OPTIMIZE_AND
  if (a==b) return a;
  if (a==~b) return mkfalse();
#endif

  int ca = toInt(a), cb = toInt(b);
#ifndef DO_NOT_REORDER_AND
  if (ca > cb) {
    int z = ca;
    ca = cb;
    cb = z;
  }
#endif
  std::pair<int, int> idx(ca, cb);
  
  std::map <std::pair<int, int>, Var>::iterator pos = andMap.find(idx);
  
  if (pos == andMap.end()) {
    Var v = newVar();
    andMap.insert(pos, std::pair<std::pair<int, int>, Var>(idx, v));
    
    Lit o(v);
    vec<Lit> clause1, clause2, clause3;
    
    clause1.push(~a);
    clause1.push(~b);
    clause1.push(o);
    addClause(clause1);
    
    clause2.push(a);
    clause2.push(~o);
    addClause(clause2);
    
    clause3.push(b);
    clause3.push(~o);
    addClause(clause3);
    
    return o;
  } else {
    return Lit(pos->second);
  }
}

Lit LinearSMT::linearAtom(const Constraint& c) {
  Linear_Expression e(c);
  if (c.is_equality()) {
    return(mkand(exprAtom(e), exprAtom(-e)));
    }
  else if (c.is_strict_inequality()) {
    return ~exprAtom(-e);
  } else {
    return exprAtom(e);
  }
};

std::ostream& operator<<(std::ostream& o, lbool v) {
  switch (v.toInt()) {
  case 1:
    o << "true";
    break;
  case -1:
    o << "false";
    break;
  case 0:
    o << "undef";
    break;
  }
  return o;
}

bool sameDirection (const PPL::Linear_Expression& a, const PPL::Linear_Expression& b) {
  unsigned m = max(a.space_dimension(), b.space_dimension());
  for(unsigned i=0; i<m; i++) {
    mpz_class ac, bc;
    if (i < a.space_dimension()) ac = a.coefficient(Variable(i));
    if (i < b.space_dimension()) bc = b.coefficient(Variable(i));
    if (ac != bc) return false;
  }
  return true;
}

void LinearSMT::generateTrivialTheoryConsequences() {  
  /* std::cout << "generate trivial" << std::endl; 
  for (constraintMap_const_iterator i=constraintMap_begin();
       i != constraintMap_end(); ++i) {
    std::cout << i->second << std::endl;
    } */

  // Phase 1: B >= A => L(X) + A >= 0 => L(X) + B >= 0
  {
    constraintMap_const_iterator i=constraintMap_begin(), z=constraintMap_end();
    if (i!=z) {
      constraintMap_const_iterator next=i;
      ++next;
      for(; next!=z; i=next, ++next) {
	if (sameDirection(i->first, next->first)) {
#ifdef SMT_DEBUG
	std::cout << i->first << " >= 0 => " << next->first << " >= 0" << std::endl; 
#endif
	  addImplication(Lit(i->second), Lit(next->second));
	}
      }
    }
  }

  // Phase 2
  {
    for(constraintMap_const_iterator
	  i=constraintMap_begin(),
	  z = constraintMap_end();
	  i != z; i++) {
      PPL::Linear_Expression iNeg(- i->first);
      constraintMap_iterator j = constraintMap.lower_bound(iNeg);
      
      if (j != z && i->second < j->second && sameDirection(iNeg, j->first)) {
#ifdef SMT_DEBUG
	std::cout << i->first << " >= 0 || " << j->first << " >= 0" << std::endl; 
	std::cout << "nClauses=" << nClauses() << std::endl;
#endif
	addDisjunction(Lit(i->second), Lit(j->second));
      }
    }
  }
}

std::ostream& operator<<(std::ostream& o, Lit l) {
  if (sign(l)) o<<"~";
  o << var(l);
  return o;
}

void LinearSMT::addImplication(Lit a, Lit b) {
  vec<Lit> v;
  v.push(~a);
  v.push(b);
  addClause(v);
}

void LinearSMT::addDisjunction(Lit a, Lit b) {
  vec<Lit> v;
  v.push(a);
  v.push(b);
  addClause(v);
}

#ifndef USE_DIRECT_PPL
Constraint_Polyhedron::Constraint_Polyhedron(unsigned dimension) :
  problem(PPL::Constraint_System(PPL::Variable(0) >= 0), PPL::Variable(0), PPL::MAXIMIZATION) {
}

void Constraint_Polyhedron::add_constraint(const PPL::Constraint& c) {
  Linear_Expression lOrig(c), lShifted(lOrig.inhomogeneous_term());
  for(unsigned i=0; i<lOrig.space_dimension(); i++) {
    mpz_class coeff = lOrig.coefficient(PPL::Variable(i));
    if (coeff != 0) {
      lShifted += PPL::Variable(i+1) * coeff;
    }
  }
  if (c.is_equality()) {
    problem.add_constraint(lShifted == 0);
  } else if (c.is_nonstrict_inequality()) {
    problem.add_constraint(lShifted >= 0);
  } else {
    problem.add_constraint(lShifted >= PPL::Variable(0));
  }
}

bool Constraint_Polyhedron::is_empty() const {
  //std::cout << "emptyness test" << std::endl;
  switch (problem.solve()) {
  case PPL::UNFEASIBLE_LP_PROBLEM:
    //std::cout << "unfeasible" << std::endl;
    return true;
  case PPL::UNBOUNDED_LP_PROBLEM:
    //std::cout << "unbounded" << std::endl;
    return false;
  case PPL::OPTIMIZED_LP_PROBLEM:
    {
      mpz_class num, den;
      problem.optimal_value(num, den);
      //std::cout << "optimal value " << num << "/" << den << std::endl;
      return ! ((num > 0 && den > 0) || (num < 0 && den < 0));
    }
  default:
    std::terminate();
  }
}
#endif
