#include "smtlib_lexer.h"
#include "FormulaTranslator.hh"
#include <iostream>
#include <set>
#include <stdio.h>

#ifdef __GNUC__
#define noreturn __attribute__((noreturn))
#else
#define noreturn
#endif

template<class T> class reference {
  T *ptr;

public:
  T& operator*() const {
    return *ptr;
  }
  T* operator->() const {
    return ptr;
  }

  reference(T* p) : ptr(p) {}

  ~reference() { delete ptr; }
  
  bool isNull() { return ptr==0; }
};

enum sat_status { SAT, UNSAT, UNKNOWN };

struct smtlib_parser {
  enum token last_token;

  enum token get_token() {
    if (last_token == NONE) {
      last_token = smtlib_lex();
    }
    return last_token;
  }

  void next_token() {
    last_token = NONE;
  }

  void parse_formula_list(ManySubFormulas& l);
  void parse_expression_list(ManyExpression& l);

  Formula* parse_formula();
  ManySubFormulas *parse_compound_formula(ManySubFormulas* l);
  Comparison *parse_comparison(Comparison::ComparisonType kind);
  Expression *parse_expression();
  void parse_benchmark();

  Formula *formula;
  enum sat_status status;

  smtlib_parser() : last_token (NONE), formula(0), status(UNKNOWN) {
  }

  void parse_error(const std::string& msg) const noreturn;
};

void smtlib_parser::parse_error(const std::string& msg) const {
  std::cerr << "parse error at line " << smtlib_lineno << ", token ";
  switch(last_token) {
  case NONE:
    std::cerr << "<none>"; break;
  case END_OF_FILE:
    std::cerr << "<EOF>"; break;
  case LPAREN:
    std::cerr << "("; break;
  case RPAREN:
    std::cerr << ")"; break;
  case IDENTIFIER:
    std::cerr << "identifier[" << smtlib_text << "]"; break;
  case INTEGER:
    std::cerr << "integer[" << smtlib_text << "]"; break;
  case LESS:
    std::cerr << "<"; break;
  case LESS_EQ:
    std::cerr << "<="; break;
  case GREATER:
    std::cerr << ">"; break;
  case GREATER_EQ:
    std::cerr << ">="; break;
  case AND:
    std::cerr << "'and'"; break;
  case OR:
    std::cerr << "'or'"; break;
  case NOT:
    std::cerr << "'not'"; break;
  case PLUS:
    std::cerr << "+"; break;
  case MINUS:
    std::cerr << "-"; break;
  case NEG:
    std::cerr << "~"; break;
  case MULTIPLY:
    std::cerr << "*"; break;
  case DIVIDE:
    std::cerr << "/"; break;
  case FLET:
    std::cerr << "flet"; break;
  case LET:
    std::cerr << "let"; break;
  case EQUAL:
    std::cerr << "="; break;
  case DISTINCT:
    std::cerr << "distinct"; break;
  case IF_THEN_ELSE:
    std::cerr << "if_then_else"; break;
  case ITE:
    std::cerr << "ite"; break;
  case IFF:
    std::cerr << "iff"; break;
  case BENCHMARK:
    std::cerr << "benchmark"; break;
  case COLON:
    std::cerr << ":"; break;
  case STRING:
    std::cerr << "<string>"; break;
  case KW_SAT:
    std::cerr << "sat"; break;
  case KW_UNSAT:
    std::cerr << "unsat"; break;
  case KW_UNKNOWN:
    std::cerr << "unknown"; break;
  case STATUS:
    std::cerr << "status"; break;
  case SOURCE:
    std::cerr << "source"; break;
  case CATEGORY:
    std::cerr << "category"; break;
  case DIFFICULTY:
    std::cerr << "difficulty"; break;
  case LOGIC:
    std::cerr << "logic"; break;
  case EXTRAPREDS:
    std::cerr << "extrapreds"; break;
  case EXTRAFUNS:
    std::cerr << "extrafuns"; break;
  case FORMULA:
    std::cerr << "formula"; break;
  case NOTES:
    std::cerr << "notes"; break;
  }
  std::cerr << " : " << msg << std::endl;
  exit(1);
}

void smtlib_parser::parse_formula_list(ManySubFormulas& l) {
  while(true) {
    Formula *f = parse_formula();
    if (f == 0) break;
    l.addFormula(f);
  }
}

ManySubFormulas* smtlib_parser::parse_compound_formula(ManySubFormulas *f) {
  next_token();
  parse_formula_list(*f);
  if (get_token() != RPAREN) parse_error("expecting formula in list of formulas");
  next_token();
  return f;
}

Expression* smtlib_parser::parse_expression() {
  switch (get_token()) {
  case IDENTIFIER:
    next_token();
    return new NumericVariable(smtlib_text);

  case INTEGER:
    next_token();
    return new IntegerConstant(mpz_class(smtlib_text));

  case LPAREN:
    {
      next_token();
      Expression *o;
      switch (get_token()) {
      case NEG:
	{
	  next_token();
	  Expression *f = parse_expression();
	  if (f==0) parse_error("expecting expression in neg");
	  o = new UnaryNeg(f);
	}
	break;

      case PLUS:
	{
	  next_token();
	  Sum *s = new Sum();
	  while (true) {
	    Expression *p = parse_expression();
	    if (p==0) break;
	    s -> addExpression(p);
	  }
	  o = s;
	}
	break;

      case MULTIPLY:
	{
	  next_token();
	  Product *s = new Product();
	  while (true) {
	    Expression *p = parse_expression();
	    if (p==0) break;
	    s -> addExpression(p);
	  }
	  o = s;
	}
	break;

      case MINUS:
	{
	  next_token();
	  Sum *s = new Sum();
	  s -> addExpression(parse_expression());
	  while (true) {
	    Expression *p = parse_expression();
	    if (p==0) break;
	    s -> addExpression(new UnaryNeg(p));
	  }
	  o = s;
	}
	break;

      case DIVIDE:
	{
	  next_token();
	  Expression *a = parse_expression();
	  if (a==0) parse_error("division: missing numerator");
	  Expression *b = parse_expression();
	  if (b==0) parse_error("division: missing denominator");
	  o = new Quotient(a, b);
	}
	break;

      case ITE:
	{
	  next_token();
	  Formula *c = parse_formula();
	  if (c==0) parse_error("ite: missing condition");
	  Expression* e1 = parse_expression();
	  if (e1==0) parse_error("ite: missing 'then' expression");
	  Expression* e2 = parse_expression();
	  if (e2==0) parse_error("ite: missing 'else' expression");
	  o = new ExpressionIfThenElse(c, e1, e2);
	}
	break;

      default:
	parse_error("expression: unknown operator");
      }
      if (get_token() != RPAREN) parse_error("expression: missing closing parenthesis");
      next_token();
      return o;
    }

  default:
    return 0;
  }
}

void smtlib_parser::parse_expression_list(ManyExpression& f) {
  while (true) {
    Expression *e = parse_expression();
    if (e==0) break;
    f.addExpression(e);
  }
}

Comparison* smtlib_parser::parse_comparison(Comparison::ComparisonType kind) {
  next_token();
  Expression* a1 = parse_expression();
  if (a1 == 0) parse_error("comparison: missing first expression");
  Expression* a2 = parse_expression();
  if (a2 == 0) parse_error("comparison: missing second expression");
  if (get_token() != RPAREN) parse_error("comparison: missing closing parenthesis");
  next_token();
  return new Comparison(a1, kind, a2);
}

Formula* smtlib_parser::parse_formula() {
  switch (get_token()) {
  case IDENTIFIER:
    next_token();
    return new PropositionalVariable(smtlib_text);

  case LPAREN:
    next_token();
    switch (get_token()) {
    case EQUAL:
      {
	next_token();
	Equality *f = new Equality();
	parse_expression_list(*f);
	if (get_token() != RPAREN) parse_error("equal: missing closing parenthesis");
	next_token();
	return f;
      }

    case DISTINCT:
      {
	next_token();
	PairwiseDistinct *f = new PairwiseDistinct();
	parse_expression_list(*f);
	if (get_token() != RPAREN) parse_error("distinct: missing closing parenthesis");
	next_token();
	return f;
      }

    case FLET:
      {
	next_token();
	if (get_token() != LPAREN) parse_error("flet: missing opening parenthesis");
	next_token();
	if (get_token() != IDENTIFIER) parse_error("flet: missing identifier");
	std::string varName = smtlib_text;
	next_token();

	Formula *binding = parse_formula();
	if (binding == 0) parse_error("flet: missing bound formula");
	if (get_token() != RPAREN) parse_error("flet: missing closing parenthesis before subformula");
	next_token();
	Formula *f = parse_formula();
	if (f == 0) parse_error("flet: missing subformula");
	if (get_token() != RPAREN) parse_error("flet: missing closing parenthesis after subformula");
	next_token();
	return new FLet(varName, binding, f);
      }

    case LET:
      {
	next_token();
	if (get_token() != LPAREN) parse_error("let: missing opening parenthesis");
	next_token();
	if (get_token() != IDENTIFIER) parse_error("let: missing identifier");
	std::string varName = smtlib_text;
	next_token();

	Expression *binding = parse_expression();
	if (binding == 0) parse_error("let: missing bound expression");
	if (get_token() != RPAREN) parse_error("let: missing closing parenthesis before subformula");
	next_token();
	Formula *f = parse_formula();
	if (f == 0) parse_error("let: missing subformula");
	if (get_token() != RPAREN) parse_error("let: missing closing parenthesis before subformula");
	next_token();
	return new ExprLet(varName, binding, f);
      }

    case NOT:
      {
	next_token();
	Formula *f = parse_formula();
	if (f == 0 || get_token() != RPAREN) parse_error("not: missing closing parenthesis");
	next_token();
	return new Not(f);
      }

    case IF_THEN_ELSE:
      {
	next_token();
	Formula *c = parse_formula();
	if (c==0) parse_error("if_then_else: missing condition");
	Formula *e1 = parse_formula();
	if (e1==0) parse_error("if_then_else: missing 'then' subformula");
	Formula *e2 = parse_formula();
	if (e2==0 || get_token() != RPAREN) parse_error("if_then_else: missing 'else' subformula");
	next_token();
	return new FormulaIfThenElse(c, e1, e2);
      }
      
    case AND:
      return parse_compound_formula(new And());

    case OR:
      return parse_compound_formula(new Or());

    case IFF:
      return parse_compound_formula(new Iff());

    case LESS_EQ:
      return parse_comparison(Comparison::LESS_EQ);

    case LESS:
      return parse_comparison(Comparison::LESS);

    case GREATER_EQ:
      return parse_comparison(Comparison::GREATER_EQ);

    case GREATER:
      return parse_comparison(Comparison::GREATER);

    default:
      parse_error("formula: unknown operator");
    }
  default:
    return 0;
  }
}

void smtlib_parser::parse_benchmark() {
  if (get_token() != LPAREN) parse_error("benchmark: opening parenthesis");
  next_token();
  if (get_token() != BENCHMARK) parse_error("benchmark: missing 'benchmark'");
  next_token();
  if (get_token() != IDENTIFIER) parse_error("benchmark: missing name");
  next_token();

  while (get_token() != RPAREN) {
    if (get_token() != COLON) parse_error("benchmark: missing colon prefix");
    next_token();
    switch (get_token()) {
    case SOURCE:
      next_token();
      if (get_token() != STRING) parse_error("source: missing string");
      next_token();
      break;

    case DIFFICULTY:
      next_token();
      if (get_token() != STRING) parse_error("difficulty: missing string");
      next_token();
      break;

    case CATEGORY:
      next_token();
      if (get_token() != STRING) parse_error("category: missing string");
      next_token();
      break;

    case NOTES:
      next_token();
      if (get_token() != STRING) parse_error("notes: missing string");
      next_token();
      break;

    case LOGIC:
      next_token();
      if (get_token() != IDENTIFIER) parse_error("benchmark: missing logic name");
      next_token();
      break;

    case STATUS:
      next_token();
      switch (get_token()) {
      case KW_SAT:
	status = SAT;
	break;
      case KW_UNSAT:
	status = UNSAT;
	break;
      case KW_UNKNOWN:
	status = UNKNOWN;
	break;
      default:
	parse_error("benchmark: missing status");
      }
      next_token();
      break;

    case EXTRAPREDS:
      next_token();
      if (get_token() != LPAREN) parse_error("extrapreds: missing opening parenthesis(1)");
      next_token();
      if (get_token() != LPAREN) parse_error("extrapreds: missing opening parenthesis(2)");
      next_token();
      if (get_token() != IDENTIFIER) parse_error("extrapreds: missing identifier");
      next_token();
      if (get_token() != RPAREN) parse_error("extrapreds: missing closing parenthesis(1)");
      next_token();
      if (get_token() != RPAREN) parse_error("extrapreds: missing closing parenthesis(2)");
      next_token();
      break;

    case EXTRAFUNS:
      next_token();
      if (get_token() != LPAREN) parse_error("extrafuns: missing opening parenthesis(1)");
      next_token();
      if (get_token() != LPAREN) parse_error("extrafuns: missing opening parenthesis(2)");
      next_token();
      if (get_token() != IDENTIFIER) parse_error("extrafuns: missing identifier(1)");
      next_token();
      if (get_token() != IDENTIFIER) parse_error("extrafuns: missing identifier(2)");
      next_token();
      if (get_token() != RPAREN) parse_error("extrafuns: missing closing parenthesis(1)");
      next_token();
      if (get_token() != RPAREN) parse_error("extrafuns: missing closing parenthesis(2)");
      next_token();
      break;

    case FORMULA:
      next_token();
      {
	formula = parse_formula();
	if (formula == 0) parse_error("benchmark: missing formula");
      }
      break;

   default:
      parse_error("benchmark: unknown tag");
    }
  }
  next_token();

  if (get_token() != END_OF_FILE) parse_error("benchmark: missing end of file");
}

int main(int argc, char **argv) {
  if (argc != 2) {
    std::cerr << "usage: " << argv[0] << " file.smt" << std::endl;
    exit(1);
  }

  smtlib_parser parser;
  smtlib_in = fopen(argv[1], "r");
  parser.parse_benchmark();

  reference <Formula> f (parser.formula);

  if (f.isNull()) {
    std::cerr << "bad formula" << std::endl;
    exit(1);
  }

#if 0
  std::cout << "PROPOSITIONAL VARIABLES" << std::endl;
  std::set<std::string> prop;
  f->collectPropositionalVariables(prop);
  for (std::set<std::string> ::const_iterator i=prop.begin(), end=prop.end();
       i != end; ++i) {
    std::cout << *i << std::endl;
  }

  std::cout << "NUMERIC VARIABLES" << std::endl;
  std::set<std::string> num;
  f->collectNumericVariables(num);
  for (std::set<std::string> ::const_iterator i=num.begin(), end=num.end();
       i != end; ++i) {
    std::cout << *i << std::endl;
  }
#endif

  std::cout << "generating model" << std::endl;

  FormulaTranslator solver(*f);
  if (solver.solve()) {
    solver.printModel(std::cout);
    if (parser.status == UNSAT)
      std::cerr << "PROBLEM" << std::endl;
  } else {
    std::cout << "unsolvable" << std::endl;
    if (parser.status == SAT)
      std::cerr << "PROBLEM" << std::endl;
  }

  return 0;
}
